from __future__ import annotations

from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Any, Callable, Dict, Optional

from .utils import sanitize_url


@dataclass
class ElementDescription:
    """The description of element contents.

    ## `name`
    - unique name of the element on the page

    ## `type`
    - type of the element, one of predefined types

    ## `configuration`
    - configuration values specific for each element
    """

    name: str
    type: str
    configuration: Dict


class ElementBase(ABC):
    """Base class for all elements in a single component. Element is a basic
    piece of information on the page, e.g. heading, table, selection input
    element...
    """

    def __init__(self, name: str):
        self._name = name
        self.changed = True

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value: str):
        self._name = value

    @abstractmethod
    def construct_element_description(self) -> ElementDescription:
        ...

    @abstractmethod
    def add_endpoint(self, app):
        ...


class URLNamedWrapper:
    def __init__(self, internal_object, attr_name: str = "endpoint_url"):
        self.internal_object = internal_object
        self.attr_name = attr_name

    @property
    def name(self):
        return getattr(self.internal_object, self.attr_name)

    @name.setter
    def name(self, value: str):
        setattr(self.internal_object, self.attr_name, value)


class ElementWithEndpoint(ElementBase):
    def __init__(
        self,
        name: str,
        endpoint_callback: Callable,
        endpoint_url: Optional[str] = None,
    ):
        super().__init__(name)
        if endpoint_url is None:
            endpoint_url = sanitize_url(self.name)
        self.endpoint_url = endpoint_url
        self.endpoint_callback = endpoint_callback
        self.parent_component: Optional[Any] = None
        """The component that holds all the other elements. This is set
        in `ComponentBase.register_elements`
        """

    def get_url_named_wrapper(self):
        return URLNamedWrapper(self)



from typing import (  # noqa
    Any,
    Callable,
    Dict,
    KeysView,
    List,
    Optional,
    Protocol,
    Union,
)

from visuallm.elements.plain_text_element import PlainTextElement
from visuallm.elements.selector_elements import (
    ButtonElement,
    ChoicesSubElement,
    MinMaxSubElement,
)


class DatasetProtocol(Protocol):
    def __getitem__(self, __key: str) -> List[Any]:
        ...

    def keys(self) -> KeysView[str]:
        ...


DATASET_TYPE = Union[DatasetProtocol, Callable[[], DatasetProtocol]]
DATASETS_TYPE = Union[
    Dict[str, DatasetProtocol],
    Dict[str, Callable[[], DatasetProtocol]],
]


class DataPreparationMixin:
    def __init__(
        self,
        on_sample_change_callback: Callable[[], None],
        dataset: Optional[DATASET_TYPE] = None,
        dataset_choices: Optional[DATASETS_TYPE] = None,
        keep_datasets_in_memory: bool = True,
        update_on_data_config_sent: bool = True,
    ):
        self._dataset, self._dataset_choices = None, None
        if dataset_choices is not None:
            assert len(dataset_choices) != 0
            self._dataset_choices = dataset_choices
            values = dataset_choices.values()
            self._dataset = self.load_dataset(next(iter(values)))

        if dataset is not None and dataset_choices is None:
            assert dataset_choices is None
            self._dataset_choices = None
            self._dataset = self.load_dataset(dataset)

        if keep_datasets_in_memory:
            self._dataset_cache: Optional[Dict[str, DatasetProtocol]] = {}
        else:
            self._dataset_cache = None
        self.initialize_data_button()
        self._on_sample_change_callback = on_sample_change_callback
        self._loaded_sample: Any = self.get_split()[
            int(self.sample_selector_element.selected)
        ]
        self._update_on_data_config_sent = update_on_data_config_sent

    def data_force_update(self):
        self.dataset_split_selector_element.force_set_updated()
        self.sample_selector_element.force_set_updated()
        if self.dataset_selector_element is not None:
            self.dataset_selector_element.force_set_updated()

    def load_dataset(
        self, value: Union[DatasetProtocol, Callable[[], DatasetProtocol]]
    ):
        if callable(value):
            return value()
        else:
            return value

    def load_cached_dataset(
        self, name: str, load_dataset_fn: Callable[[], DatasetProtocol]
    ):
        if self._dataset_cache is not None and name in self._dataset_cache:
            self._dataset = self._dataset_cache[name]
        else:
            self._dataset = load_dataset_fn()
            if self._dataset_cache is not None:
                self._dataset_cache[name] = self._dataset

        self.dataset_split_selector_element.set_choices(
            self.get_dataset_splits(self._dataset)
        )

    def get_dataset_splits(self, dataset: Optional[DatasetProtocol]):
        if dataset is None:
            return ["dummy_split", "dummy_split"]
        return list(dataset.keys())

    def initialize_data_button(self):
        self.dataset_selector_heading = PlainTextElement(
            content="Dataset Settings", is_heading=True
        )
        self.dataset_split_selector_element = ChoicesSubElement(
            choices=self.get_dataset_splits(self._dataset),
            text="Select Dataset Split",
        )
        self.sample_selector_element = MinMaxSubElement(
            sample_min=0,
            sample_max=len(self.get_split()) - 1,
            text="Select Dataset Sample",
        )

        subelements = [
            self.dataset_split_selector_element,
            self.sample_selector_element,
        ]
        if self._dataset_choices is not None:
            self.dataset_selector_element = ChoicesSubElement(
                list(self._dataset_choices.keys()), text="Select Dataset"
            )
            subelements.append(self.dataset_selector_element)
        else:
            self.dataset_selector_element = None

        self.dataset_button = ButtonElement(
            button_text="Send Dataset Configuration",
            processing_callback=self.dataset_callback,
            subelements=subelements,
        )

    def get_split(self) -> List[Any]:
        if self._dataset is None:
            return ["dummy_sample", "dummy_sample_2"]
        return self._dataset[self.dataset_split_selector_element.selected]

    @property
    def dataset_elements(self):
        return [self.dataset_selector_heading, self.dataset_button]

    def dataset_callback(self):
        """Check `self.sample_selector_element` and
        `self.dataset_split_selector_element` and switch all the associated
        values."""
        assert self._dataset is not None
        if self.dataset_selector_element is not None:
            if self.dataset_selector_element.updated:
                assert self._dataset_choices is not None
                key = self.dataset_selector_element.selected
                args = self._dataset_choices[key]
                self.load_cached_dataset(key, lambda: self.load_dataset(args))
        if self.dataset_split_selector_element.updated:
            self.sample_selector_element._max = len(self.get_split()) - 1
            self.sample_selector_element.selected = min(
                self.sample_selector_element._max,
                self.sample_selector_element.selected,
            )
            self.sample_selector_element.force_set_updated()
        if self.sample_selector_element.updated:
            self._loaded_sample = self.get_split()[
                int(self.sample_selector_element.selected)
            ]
            self._on_sample_change_callback()
        elif self._update_on_data_config_sent:
            self._on_sample_change_callback()



from __future__ import annotations

from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Any, Callable, Dict, Generic, List, MutableSet, Optional, TypeVar

from flask import request

from visuallm.server import Server

from .element_base import ElementDescription, ElementWithEndpoint
from .utils import register_named


@dataclass
class SubElementConfiguration:
    subtype: str
    name: str
    configuration: Dict[str, Any]
    parent_name: str


class ButtonElement(ElementWithEndpoint):
    """I expect the following flow of data:
    - In frontend the user selects some value from the selector (automatic)
    - In frontend the user clicks the button element which is the parent of
        the selector (automatic)
    - The data arrives to backend, where each selector is updated and signalizes
        that is has been updated through `self.updated` flag (automatic)
    - the programmer can control what is influenced by updated selectors
    - all the changed elements are sent back to the frontend (automatic)
    """

    def __init__(
        self,
        processing_callback: Callable[[], None],
        name: str = "selector",
        subelements: List[SelectorSubElement] = [],
        button_text="Select",
        **kwargs,
    ):
        """
        Args:
            processing_callback (Callable[[], None]): a function that is
                called just after all the data from the frontend have been
                processed.
            name (str, optional): name of the element, doesn't have to be
                provided. Defaults to "selector".
            subelements (List[SelectorSubElement], optional): input
                subelements. E.g. if there is a checkbox on each press of a
                button, the checkbox value will be sent to the backend.
                Defaults to [].
            button_text (str, optional): Text displayed in a button input
                element. Defaults to "Select".
        """
        if "endpoint_callback" not in kwargs:
            kwargs["endpoint_callback"] = self.default_select_callback
        super().__init__(name=name, **kwargs)
        self.type = "sample_selector"
        self.processing_callback = processing_callback
        self._button_text = button_text
        self._subelements_dict: Dict[str, SelectorSubElement] = {}
        self._subelements: List[SelectorSubElement] = []
        self._subelement_names: MutableSet[str] = set()

        for subelement in subelements:
            self.add_subelement(subelement)

    def construct_element_description(self):
        self.changed = False
        return ElementDescription(
            name=self.name,
            type=self.type,
            configuration=dict(
                address=self.endpoint_url.removeprefix("/"),
                button_text=self._button_text,
                subelement_configs=[
                    c.subelement_configuration for c in self._subelements
                ],
            ),
        )

    def default_select_callback(self):
        """Goes over the standard format of response from FE and sets all
        the relevant selected attributes in subelement selectors, hten returns
        the control to the programmer for handling of the updated data and
        then returns everything updated to the frontend.
        """
        if not request.is_json:
            raise RuntimeError()
        assert self.parent_component is not None
        response_json = request.get_json()
        for key, value in response_json.items():
            self._subelements_dict[key].selected = value

        self.processing_callback()
        return self.parent_component.fetch_info(fetch_all=False)

    def add_subelement(self, subelement: SelectorSubElement):
        if subelement.parent_element is not None:
            raise RuntimeError()
        subelement.parent_element = self

        register_named(subelement, self._subelement_names, self._subelements)
        self._subelements_dict[subelement.name] = subelement

    def add_endpoint(self, app: Server):
        app.add_endpoint(self.endpoint_url, self.endpoint_callback, methods=["POST"])


SelectedType = TypeVar("SelectedType")


class SelectorSubElement(ABC, Generic[SelectedType]):
    """I expect the following flow of data:
    - In frontend the user selects some value from the selector (automatic)
    - In frontend the user clicks the button element which is the parent of
        the selector (automatic)
    - The data arrives to backend, where each selector is updated and signalizes
        that is has been updated through `self.updated` flag (automatic)
    - the programmer can control what is influenced by updated selectors
    - all the changed elements are sent back to the frontend (automatic)
    """

    @property
    def subelement_configuration(self) -> SubElementConfiguration:
        if self.parent_element is None:
            raise RuntimeError()
        return SubElementConfiguration(
            self._subtype,
            self.name,
            dict(
                selected=self._selected,
                text=self._text,
                **self.construct_selector_data(),
            ),
            self.parent_element.name,
        )

    @abstractmethod
    def construct_selector_data(self) -> Dict[str, Any]:
        ...

    @property
    @abstractmethod
    def selected(self):
        ...

    @selected.setter
    @abstractmethod
    def selected(self, value: SelectedType):
        ...

    def selected_getter(self) -> SelectedType:
        assert self._selected is not None
        return self._selected

    def selected_setter(self, value: SelectedType):
        if self.parent_element is None:
            raise ValueError(
                "Cannot change the value of the element without atributing "
                + "the element to the parent component"
            )
        self._updated = value != self._selected
        self.parent_element.changed |= self._updated
        self._selected = value

    def force_set_updated(self):
        """Set updated to true, so that any changes associated with the update
        are triggered"""
        self._updated = True

    @property
    def updated(self):
        """Whether the selector was updated by the frontend."""
        return self._updated

    def __init__(self, subtype: str, text: str):
        """WARNING:
        `subtype` must match the subtype field in frontend.
        """
        self.name = str(subtype)
        self._updated = True
        self._subtype = subtype
        self._selected: Optional[SelectedType] = None
        self.parent_element: Optional[ButtonElement] = None
        self._text = text


class MinMaxSubElement(SelectorSubElement[float]):
    """Subelement in the ButtonElement that creates an int selection in
    a range. E.g. selector between [min, max].
    """

    def __init__(
        self,
        sample_min: float,
        sample_max: float,
        text: str,
        step_size: float = 1.0,
        default_value: Optional[float] = None,
    ):
        super().__init__(subtype="min_max", text=text)
        if sample_min > sample_max:
            raise ValueError(
                f"sample_min ({sample_min}) should be bigger than or equal "
                f"to sample_max ({sample_max})"
            )
        if default_value is None:
            default_value = sample_min
        self._selected: float = default_value
        self._min = sample_min
        self._max = sample_max
        self._step_size = step_size

    @property
    def selected(self):
        return self.selected_getter()

    @selected.setter
    def selected(self, value: float):
        if (value > self._max) or (value < self._min):
            raise ValueError(
                f"Invalid value to selected ({value}) should be in range: ["
                + f"{self._min}, {self._max}]"
            )
        self.selected_setter(value)

    def construct_selector_data(self) -> Dict[str, Any]:
        self._updated = False
        return dict(min=self._min, max=self._max, step_size=self._step_size)


class ChoicesSubElement(SelectorSubElement[str]):
    def __init__(self, choices: List[str], text: str):
        """Default selected value is the first value in the `choices` list"""
        super().__init__(subtype="choices", text=text)
        if len(choices) == 0:
            raise RuntimeError("choices should have length at least 1!")
        self._selected = choices[0]
        self._choices = choices

    @property
    def selected(self):
        return self.selected_getter()

    @selected.setter
    def selected(self, value: str):
        if value not in self._choices:
            raise ValueError(
                f"Invalid value to selected ({value}), "
                + f"possibilities: {self._choices}"
            )
        self.selected_setter(value)

    def set_choices(self, new_choices: List[str]):
        if len(new_choices) == 0:
            raise RuntimeError("Choices should have length at least 1!")
        self._choices = new_choices
        if self._selected not in self._choices:
            self._selected = self._choices[0]
        self.force_set_updated()

    def construct_selector_data(self) -> Dict[str, Any]:
        self._updated = False
        return dict(choices=self._choices)


class CheckBoxSubElement(SelectorSubElement[bool]):
    def __init__(self, text: str, default_value: bool = False):
        super().__init__(subtype="check_box", text=text)
        self._selected = default_value

    @property
    def selected(self):
        return self.selected_getter()

    @selected.setter
    def selected(self, value: bool):
        if not isinstance(value, bool):
            raise ValueError(f"Invalid value assigned to bool: {value}")
        self.selected_setter(value)

    def construct_selector_data(self) -> Dict[str, Any]:
        self._updated = False
        return {}
